@@ -5,13 +5,14 @@
             [clj-http.multipart :as mp]
             [clj-http.util :refer [opt]]
             [clojure.pprint])
-  (:import (java.io ByteArrayOutputStream File FilterInputStream InputStream)
-           (java.net URI)
+  (:import (java.io ByteArrayOutputStream FilterInputStream InputStream)
+
            (org.apache.http HeaderIterator HttpEntity
                             HttpEntityEnclosingRequest
                             HttpResponse Header HttpHost
                             HttpResponseInterceptor)
-           (org.apache.http.auth UsernamePasswordCredentials AuthScope)
+           (org.apache.http.auth UsernamePasswordCredentials AuthScope
+                                 NTCredentials)
            (org.apache.http.params CoreConnectionPNames)
            (org.apache.http.client HttpClient HttpRequestRetryHandler)
            (org.apache.http.client.methods HttpDelete
@@ -30,14 +31,14 @@
            (org.apache.http.impl.client DefaultHttpClient)
            (org.apache.http.impl.conn ProxySelectorRoutePlanner)
            (org.apache.http.impl.cookie BrowserCompatSpec)
-           (org.apache.http.util EntityUtils)))
+           (java.net URI)))
 
 (defn parse-headers
   "Takes a HeaderIterator and returns a map of names to values.
 
-   If a name appears more than once (like `set-cookie`) then the value
-   will be a vector containing the values in the order they appeared
-   in the headers."
+  If a name appears more than once (like `set-cookie`) then the value
+  will be a vector containing the values in the order they appeared
+  in the headers."
   [^HeaderIterator headers & [use-header-maps-in-response?]]
   (if-not use-header-maps-in-response?
     (->> (headers/header-iterator-seq headers)
@@ -59,7 +60,7 @@
 
 (defn make-proxy-method-with-body
   [method]
-  (fn [url]
+  (fn [^String url]
     (doto (proxy [HttpEntityEnclosingRequestBase] []
             (getMethod [] (.toUpperCase (name method))))
       (.setURI (URI. url)))))
@@ -134,16 +135,16 @@
     (doseq [[k v] kvs]
       (set-client-param http-client
                         k (cond
-                           (and (not= ClientPNames/CONN_MANAGER_TIMEOUT k)
-                                (instance? Long v))
-                           (Integer. ^Long v)
-                           true v)))))
+                            (and (not= ClientPNames/CONN_MANAGER_TIMEOUT k)
+                                 (instance? Long v))
+                            (Integer. ^Long v)
+                            true v)))))
 
 (defn- coerce-body-entity
   "Coerce the http-entity from an HttpResponse to either a byte-array, or a
   stream that closes itself and the connection manager when closed."
   [{:keys [as]} ^HttpEntity http-entity ^ClientConnectionManager conn-mgr]
-  (when http-entity
+  (if http-entity
     (proxy [FilterInputStream]
         [^InputStream (.getContent http-entity)]
       (close []
@@ -153,7 +154,9 @@
             (proxy-super close))
           (finally
             (when-not (conn/reusable? conn-mgr)
-              (.shutdown conn-mgr))))))))
+              (conn/shutdown-manager conn-mgr))))))
+    (when-not (conn/reusable? conn-mgr)
+      (conn/shutdown-manager conn-mgr))))
 
 (defn- print-debug!
   "Print out debugging information to *out* for a given request."
@@ -161,22 +164,22 @@
   (println "Request:" (type body))
   (clojure.pprint/pprint
    (assoc req
-     :body (if (opt req :debug-body)
-             (cond
-              (isa? (type body) String)
-              body
+          :body (if (opt req :debug-body)
+                  (cond
+                    (isa? (type body) String)
+                    body
 
-              (isa? (type body) HttpEntity)
-              (let [baos (ByteArrayOutputStream.)]
-                (.writeTo ^HttpEntity body baos)
-                (.toString baos "UTF-8"))
+                    (isa? (type body) HttpEntity)
+                    (let [baos (ByteArrayOutputStream.)]
+                      (.writeTo ^HttpEntity body baos)
+                      (.toString baos "UTF-8"))
 
-              :else nil)
-             (if (isa? (type body) String)
-               (format "... %s bytes ..."
-                       (count body))
-               (and body (bean body))))
-     :body-type (type body)))
+                    :else nil)
+                  (if (isa? (type body) String)
+                    (format "... %s bytes ..."
+                            (count body))
+                    (and body (bean body))))
+          :body-type (type body)))
   (println "HttpRequest:")
   (clojure.pprint/pprint (bean http-req)))
 
@@ -204,14 +207,15 @@
 
 (defn request
   "Executes the HTTP request corresponding to the given Ring request map and
-   returns the Ring response map corresponding to the resulting HTTP response.
+  returns the Ring response map corresponding to the resulting HTTP response.
 
-   Note that where Ring uses InputStreams for the request and response bodies,
-   the clj-http uses ByteArrays for the bodies."
+  Note that where Ring uses InputStreams for the request and response bodies,
+  the clj-http uses ByteArrays for the bodies."
   [{:keys [request-method scheme server-name server-port uri query-string
            headers body multipart socket-timeout conn-timeout proxy-host
-           proxy-ignore-hosts proxy-port as cookie-store retry-handler
-           response-interceptor digest-auth connection-manager client-params]
+           proxy-ignore-hosts proxy-port proxy-user proxy-pass as cookie-store
+           retry-handler response-interceptor digest-auth ntlm-auth
+           connection-manager client-params]
     :as req}]
   (let [^ClientConnectionManager conn-mgr
         (or connection-manager
@@ -241,6 +245,17 @@
        (.getCredentialsProvider http-client)
        (AuthScope. nil -1 nil)
        (UsernamePasswordCredentials. user pass)))
+    (when-let [[user password host domain] ntlm-auth]
+      (.setCredentials
+       (.getCredentialsProvider http-client)
+       (AuthScope. nil -1 nil)
+       (NTCredentials. user password host domain)))
+
+    (when (and proxy-user proxy-pass)
+      (let [authscope (AuthScope. proxy-host proxy-port)
+            creds (UsernamePasswordCredentials. proxy-user proxy-pass)]
+        (.setCredentials (.getCredentialsProvider http-client)
+                         authscope creds)))
     (let [http-url (str scheme "://" server-name
                         (when server-port (str ":" server-port))
                         uri
@@ -295,20 +310,20 @@
                            #(if (opt req :debug-body)
                               (assoc % :body-content
                                      (cond
-                                      (isa? (type (:body %)) String)
-                                      (:body %)
+                                       (isa? (type (:body %)) String)
+                                       (:body %)
 
-                                      (isa? (type (:body %)) HttpEntity)
-                                      (let [baos (ByteArrayOutputStream.)]
-                                        (.writeTo ^HttpEntity (:body %) baos)
-                                        (.toString baos "UTF-8"))
+                                       (isa? (type (:body %)) HttpEntity)
+                                       (let [baos (ByteArrayOutputStream.)]
+                                         (.writeTo ^HttpEntity (:body %) baos)
+                                         (.toString baos "UTF-8"))
 
-                                      :else nil))
+                                       :else nil))
                               %))
                 (assoc-in [:request :http-req] http-req)
                 (dissoc :save-request?))
             resp))
         (catch Throwable e
           (when-not (conn/reusable? conn-mgr)
-            (.shutdown ^ClientConnectionManager conn-mgr))
+            (conn/shutdown-manager conn-mgr))
           (throw e))))))
