@@ -1,30 +1,55 @@
 (ns puppetlabs.puppetdb.jdbc
   "Database utilities"
-  (:import (com.zaxxer.hikari HikariDataSource HikariConfig)
-           [java.sql Connection SQLException]
-           (java.util.concurrent TimeUnit))
   (:require [clojure.java.jdbc :as sql]
-            [clojure.string :as string]
-            [clojure.tools.logging :as log]
-            [puppetlabs.kitchensink.core :as kitchensink]
             [clojure.string :as str]
+            [clojure.tools.logging :as log]
+            [honey.sql :as hsql]
+            [puppetlabs.kitchensink.core :as kitchensink]
             [puppetlabs.puppetdb.time :as pl-time]
             [puppetlabs.puppetdb.jdbc.internal :refer [limit-result-set!]]
             [puppetlabs.puppetdb.schema :as pls :refer [defn-validated]]
             [schema.core :as s]
-            [clojure.math.numeric-tower :as math]
-            [puppetlabs.i18n.core :refer [trs]]))
+            [puppetlabs.i18n.core :refer [trs]])
+  (:import
+   (com.zaxxer.hikari HikariDataSource HikariConfig)
+   (java.sql Connection SQLException SQLTransientConnectionException)
+   (org.postgresql.util PGobject)))
 
 (def ^:dynamic *db* nil)
 
 (defn db []
   *db*)
 
+(defn sql-state [kw-name]
+  (or ({:admin-shutdown "57P01"
+        :invalid-regular-expression "2201B"
+        :program-limit-exceeded "54000"
+        :lock-not-available "55P03"
+        :query-canceled "57014"}
+       kw-name)
+      (throw (IllegalArgumentException.
+              (trs "Requested unknown SQL state")))))
+
 (defmacro with-db-connection [spec & body]
   `(sql/with-db-connection [db# ~spec]
      (binding [*db* db#]
        ~@body)))
 
+(defmacro with-monitored-db-connection [spec status & body]
+  `(let [thread# (Thread/currentThread)
+         status# ~status]
+     (locking status#
+       (swap! status# assoc :thread thread# :connecting? true))
+     (try
+       (sql/with-db-connection [db# ~spec]
+         (locking status#
+           (swap! status# dissoc :connecting?))
+         (binding [*db* db#]
+           ~@body))
+       (finally
+         (locking status#
+           (swap! status# dissoc :thread :connecting?))))))
+
 (defmacro with-db-transaction [opts & body]
   `(sql/with-db-transaction [db# (db) (hash-map ~@opts)]
      (binding [*db* db#]
@@ -32,21 +57,28 @@
 
 (defn do-commands
   "Runs the given commands in a transaction on the database given
-  by (jdbc/db).  If a command is a seq, converts it to a string
+  by (jdbc/db).  If a command is a collection, converts it to a string
   via (clojure.string/join command)."
   [& commands]
   (sql/db-do-commands *db* true
-                      (mapv #(cond
-                               (string? %) %
-                               (seq %) (string/join %)
-                               :else %)
+                      (mapv #(if (coll? %) (str/join %) %)
                             commands)))
 
 (defn do-prepared
-  "Executes an optionally parametrized sql expression in a transaction on the
-  database given by (jdbc/db)."
+  "Executes an optionally parametrized sql expression in a transaction
+  on the database given by (jdbc/db).  If a command is a collection,
+  converts it to a string via (clojure.string/join command)."
   [sql & params]
-  (sql/db-do-prepared *db* true (into [sql] params) {:multi? true}))
+  (sql/db-do-prepared *db* true
+                      (apply vector
+                             (if (coll? sql) (str/join sql) sql)
+                             params)
+                      {:multi? true}))
+
+(defn do-hsql
+  [honeysql-command]
+  (sql/db-do-prepared *db* true
+                      (hsql/format honeysql-command)))
 
 (defn do-commands-outside-txn [& commands]
   (let [^Connection conn (:connection *db*)
@@ -72,9 +104,8 @@
 (defn- insert-multi-row
   "Given a table and a list of columns, followed by a list of column
   value sequences, return a vector of the SQL needed for the insert
-  followed by the list of column value sequences. The entities
-  function specifies how column names are transformed."
-  [table columns values {:keys [on-conflict] :as opts}]
+  followed by the list of column value sequences."
+  [table columns values {:keys [on-conflict]}]
   (let [nc (count columns)
         vcs (map count values)]
     (if (not (and (or (zero? nc) (= nc (first vcs))) (apply = vcs)))
@@ -94,9 +125,8 @@
 
 (defn- insert-single-row-on-conflict
   "Given a table and a map representing a row, return a vector of the
-  SQL needed for the insert followed by the list of column values. The
-  entities function specifies how column names are transformed."
-  [table row entities {:keys [on-conflict] :as opts}]
+  SQL needed for the insert followed by the list of column values."
+  [table row {:keys [on-conflict]}]
   (let [ks (keys row)]
     (into [(str "INSERT INTO " (name table) " ( "
                 (str/join ", " (map (fn [col] (name col)) ks))
@@ -112,11 +142,11 @@
   names, a sequence of vectors of column values, one per row, and an
   options map, insert the rows into the database."
   [db table cols values opts]
-  (let [{:keys [entities transaction?]} (merge {:entities identity :transaction? true}
-                                               (when (map? db) db)
-                                               opts)
+  (let [{:keys [transaction?]} (merge {:transaction? true}
+                                      (when (map? db) db)
+                                      opts)
         sql-params (insert-multi-row table cols values opts)]
-    (if-let [con (sql/db-find-connection db)]
+    (if (sql/db-find-connection db)
       (sql/db-do-prepared db transaction? sql-params {:multi? true})
       (with-open [con (sql/get-connection db)]
         (sql/db-do-prepared (sql/add-connection db con) transaction?
@@ -141,20 +171,21 @@
     (multi-insert-helper db stmts opts)))
 
 (defn- insert-rows!
-  "Given a database connection, a table name, a sequence of rows, and
-  an options map, insert the rows into the database."
-  [db table rows {:keys [on-conflict] :as opts}]
-  (let [{:keys [entities identifiers qualifier transaction?]}
-        (merge {:entities identity :identifiers str/lower-case :transaction? true}
+  "Given a database connection, a table name, a sequence of rows, and an
+  options map, insert the rows into the database.  Options include
+  on-conflict."
+  [db table rows opts]
+  (let [{:keys [identifiers qualifier transaction?]}
+        (merge {:identifiers str/lower-case :transaction? true}
                (when (map? db) db)
                opts)
         sql-params (map (fn [row]
                           (when-not (map? row)
                             (throw (IllegalArgumentException.
                                     "insert / insert-multi! called with a non-map row")))
-                          (insert-single-row-on-conflict table row entities opts))
+                          (insert-single-row-on-conflict table row opts))
                         rows)]
-    (if-let [con (sql/db-find-connection db)]
+    (if (sql/db-find-connection db)
       (insert-helper db transaction? sql-params
                      {:identifiers identifiers :qualifier qualifier})
       (with-open [con (sql/get-connection db)]
@@ -190,15 +221,6 @@
 
 (defn query
   "Calls clojure.jdbc/query after adding (jdbc/db) as the first argument."
-  {:arglists '([sql-and-params
-                :as-arrays? false :identifiers clojure.string/lower-case
-                :result-set-fn doall :row-fn identity]
-               [sql-and-params
-                :as-arrays? true :identifiers clojure.string/lower-case
-                :result-set-fn vec :row-fn identity]
-               [[sql-string & params]]
-               [[stmt & params]]
-               [[option-map sql-string & params]])}
   [sql-params & remainder]
   (apply sql/query *db* sql-params remainder))
 
@@ -232,7 +254,7 @@
          (valid-jdbc-query? (:results-query %)))))
 
 (defn sql-array? [x]
-  (isa? (class x) java.sql.Array))
+  (instance? java.sql.Array x))
 
 (defn convert-any-sql-array [x convert]
   (if-not (sql-array? x)
@@ -252,14 +274,12 @@
   true, the result rows will be vectors, not maps, and the first
   result row will be a vector of column names."
   ([query f] (call-with-query-rows query {} f))
-  ([[sql & params]
-    {:keys [as-arrays? identifiers qualifier read-columns] :as opts}
-    f]
+  ([[sql & params] {:keys [fetch-size] :as opts} f]
    (with-db-transaction []
      (with-open [stmt (.prepareStatement ^Connection (:connection *db*) sql)]
-       (doall (map-indexed (fn [i param] (.setObject stmt (inc i) param))
-                           params))
-       (.setFetchSize stmt 500)
+       (doseq [[i param] (map vector (range) params)]
+         (.setObject stmt (inc i) param))
+       (.setFetchSize stmt (or fetch-size 500))
        (with-open [rset (.executeQuery stmt)]
          (try
            (f (sql/result-set-seq rset opts))
@@ -283,13 +303,11 @@
   call-with-query-rows, and apply any necessary conversions directly
   to each row.  (Most columns cannot be arrays.)"
   ([query f] (call-with-array-converted-query-rows query {} f))
-  ([[sql & params]
-    {:keys [as-arrays? identifiers qualifier read-columns] :as opts}
-    f]
+  ([[sql & params] {:keys [as-arrays?] :as opts} f]
    (with-db-transaction []
      (with-open [stmt (.prepareStatement ^Connection (:connection *db*) sql)]
-       (doall (map-indexed (fn [i param] (.setObject stmt (inc i) param))
-                           params))
+       (doseq [[i param] (map vector (range) params)]
+         (.setObject stmt (inc i) param))
        (.setFetchSize stmt 500)
        (let [fix-vals (if as-arrays?
                         #(mapv any-sql-array->vec %)
@@ -349,10 +367,12 @@
   "Given a list of legal result columns and a map containing a single order_by term,
   return the SQL string representing this term for use in an ORDER BY clause."
   [[field order]]
-  {:pre [(keyword? field)
+  {:pre [(or (keyword? field) (and (seq field) (keyword? (second field))))
          (contains? #{:ascending :descending} order)]
    :post [(string? %)]}
-  (str (name field)
+  (str (if (keyword? field)
+         (name field)
+         (throw (IllegalArgumentException. (trs "Non-keyword order-by term {0}" (pr-str field)))))
        (when (= order :descending) " DESC")))
 
 (defn escape-single-quotes
@@ -375,6 +395,32 @@
   [s]
   (str "\"" (escape-double-quotes s) "\""))
 
+(defn string->text-array-literal
+  "Escape string s for inclusion in a postgres text[] literal,
+  e.g. \"foo\\\"bar\" becomes the \"foo\\\\\\\"bar\" in
+  '{\"foo\\\\\\\"bar\"}'"
+  ;; https://www.postgresql.org/docs/11/arrays.html#ARRAYS-INPUT
+  ;; https://www.postgresql.org/docs/11/sql-syntax-lexical.html#SQL-SYNTAX-CONSTANTS
+  [s]
+  (assert (string? s))
+  (str \" (str/replace s "\\" "\\\\") \"))
+
+(defn str-vec->array-literal
+  "Returns a properly quoted sql values literal representing strvecs
+  as a row set of text[], e.g. [\"x\" ...] -> (array['x', ...])."
+  [strvec]
+  (str "(array[" (str/join ", " (map single-quote strvec)) "])"))
+
+(defn strs->db-array
+  [strs]
+  (assert (every? string? strs))
+  ;; https://www.postgresql.org/docs/11/arrays.html#ARRAYS-INPUT
+  (let [quoted (map string->text-array-literal strs)]
+    (doto (PGobject.)
+      (.setType "text[]")
+      (.setValue (str \{ (str/join \, quoted) \})))))
+
+;; Q: move/replace?
 (defn create-json-path-extraction
   "Given a base json field and a path of keys to traverse, construct the proper
   SQL query of the form base->'key'->'key2'..."
@@ -386,6 +432,9 @@
                ;; prefix for the first arrow in field->'key1'...
                (str "->")))))
 
+;; TODO this is no longer used by the main query engine (engine.clj), but it is
+;; still used by the legacy query engines and should be removed once it is
+;; handled by honeysql everywhere
 (pls/defn-validated paged-sql :- String
   "Given a sql string and a map of paging options, return a modified SQL string
   that contains the necessary LIMIT/OFFSET/ORDER BY clauses.  The map of paging
@@ -407,7 +456,7 @@
          (every? kitchensink/order-by-expr? order_by)]}
   (str sql
        (when-let [order-by (seq (map order-by-term->sql order_by))]
-         (str " ORDER BY " (string/join ", " order-by)))
+         (str " ORDER BY " (str/join ", " order-by)))
        (when limit (str " LIMIT " limit))
        (when offset (str " OFFSET " offset))))
 
@@ -436,82 +485,72 @@
       first
       :c))
 
-(pls/defn-validated exponential-sleep!
-  "Sleeps for a period of time, based on an adjustable base exponential backoff.
-
-   In most cases a base of 2 is sufficient, but you can adjust this to create
-   tighter or looser sleep cycles."
-  [current-attempt :- s/Int
-   base :- (s/cond-pre s/Int Double)]
-  (let [sleep-ms (-> (math/expt base current-attempt)
-                     (- 1)
-                     (* 1000))]
-    (Thread/sleep sleep-ms)))
-
-(pls/defn-validated retry-sql-or-fail :- Boolean
-  "Log the attempts made, and the final failure during SQL retries.
-
-   If there are still retries to perform, returns false."
-  [remaining :- s/Int
-   current :- s/Int
-   ^SQLException exception]
-  (cond
-   (zero? remaining)
-   (do
-     (log/warn (trs "Caught exception. Last attempt, throwing exception."))
-     (throw exception))
-
-   :else
-   (do
-     (log/debug (trs "Caught {0}: ''{1}''. SQL Error code: ''{2}''. Attempt: {3} of {4}."
-                     (.getName (class exception))
-                     (.getMessage exception)
-                     (.getSQLState exception)
-                     (inc current)
-                     (+ current remaining)))
-     (exponential-sleep! current 1.3)
-     false)))
-
-(pls/defn-validated retry-sql*
-  "Invokes (f) up to n times, retrying only if a transient connection
-  exception occurs.  The transient exceptions will be suppressed, and
-  all others will be thrown.  If the final invocation results in a
-  transient exception, it will also be thrown."
-  [n :- s/Int
-   f]
-  (loop [r n
-         current 0]
-    (if-let [result (try
-                      [(f)]
-                      ;; Catch connection errors, and retry for some of them.
-                      ;; cf. PostgreSQL docs: Appendix A. PostgreSQL Error Codes
-                      (catch java.sql.SQLTransientConnectionException e
-                        (retry-sql-or-fail r current e)))]
-      (result 0)
-      (recur (dec r) (inc current)))))
-
-(defmacro retry-sql
-  "Executes body. If a retryable error state is thrown, will retry. At most n
-   retries are done. If still some exception is thrown it is bubbled upwards in
-   the call chain."
-  [n & body]
-  `(retry-sql* ~n (fn [] ~@body)))
+(defn retry-sql
+  "Tries (f) up to max-attempts times, ignoring \"transient\"
+  exceptions (e.g. connection exceptions).  When a
+  cancellation-timeout is provided, also ignores cancelation
+  exceptions (e.g. a statement_timeout).  Throws *any* exception from
+  the final attempt."
+  [max-attempts cancellation-timeout f]
+  (let [canceled (sql-state :query-canceled)
+        attempt #(try
+                   [(f)]
+                   (catch SQLTransientConnectionException ex
+                     ex)
+                   (catch SQLException ex
+                     (when-not cancellation-timeout
+                       (throw ex))
+                     ;; canceled should include statement_timeout
+                     (when-not (= canceled (.getSQLState ex))
+                       (throw ex))
+                     ex))]
+    (loop [result (attempt)
+           i 1]
+      (cond
+        (vector? result) (result 0)
+        (>= i max-attempts)
+        (do
+          (log/warn (trs "Caught exception. Last attempt, throwing exception."))
+          (throw result))
+        :else
+        (let [ex result]
+          (assert (instance? SQLException ex))
+          (if (= canceled (.getSQLState ex))
+            (log/debug (trs "Timed out after {1}ms. Attempt: {2} of {3}."
+                            cancellation-timeout i max-attempts))
+            (log/debug (trs "Caught {0}: ''{1}''. SQL Error code: ''{2}''. Attempt: {3} of {4}."
+                            (.getName (class ex)) (.getMessage ex) (.getSQLState ex)
+                            i max-attempts)))
+          ;; i.e. approx sleeps of: 0 300 690 1197 1856 ...
+          (Thread/sleep (* (dec (Math/pow 1.3 (dec i)))
+                           1000))
+          (recur (attempt) (inc i)))))))
 
 (defn with-transacted-connection-fn
   "Calls f within a transaction with the specified clojure.jdbc isolation level.
   If isolation is nil, the connection pool default (read committed) is used.
   Retries the transaction up to 5 times."
   [db-spec isolation f]
-  ;; We've set up the connection pool to have read-committed isolation by
-  ;; default; don't explicitly specify it, as that can lead to redundant
-  ;; round-trips
-  (let [isolation (if (= :read-committed isolation)
-                    nil
-                    isolation)]
+  ;; Avoid unnecessary round-trips, when the isolation is the
+  ;; connection pool default.
+  (let [isolation (when-not (= :read-committed isolation) isolation)]
+    (retry-sql 5 nil #(with-db-connection db-spec
+                        (with-db-transaction [:isolation isolation]
+                          (f))))))
+
+(defn retry-with-monitored-connection
+  [db-spec status {:keys [isolation statement-timeout]} f]
+  ;; Avoid unnecessary round-trips, when the isolation is the
+  ;; connection pool default.
+  (let [isolation (when-not (= :read-committed isolation) isolation)]
     (retry-sql 5
-               (with-db-connection db-spec
-                 (with-db-transaction [:isolation isolation]
-                   (f))))))
+               statement-timeout
+               #(with-monitored-db-connection db-spec status
+                  (with-db-transaction [:isolation isolation]
+                    (some->> statement-timeout
+                             (format "set local statement_timeout = %d")
+                             (sql/execute! *db*))
+                    (f))))))
 
 (defmacro with-transacted-connection'
   "Executes the body within a transaction with the specified clojure.jdbc
@@ -594,8 +633,9 @@
             read-only?
             pool-name
             maximum-pool-size
-            expected-schema]
-     :as db-spec}
+            expected-schema
+            rewrite-batched-inserts]
+     :as _db-spec}
     metrics-registry]
    (let [conn-lifetime-ms (some-> conn-max-age pl-time/to-millis)
          conn-max-age-ms (some-> conn-lifetime pl-time/to-millis)
@@ -613,9 +653,17 @@
        ;; level to fail (because the pending transaction already
        ;; includes that select).
        (.setIsolateInternalQueries true))
-     (some->> (when expected-schema
-                (block-on-schema-mismatch expected-schema))
-              (.setConnectionInitSql config))
+     (->> ["DO $$ BEGIN"
+           "IF CAST((SELECT setting FROM pg_settings WHERE name = 'server_version_num') AS INTEGER) >= 110000 THEN"
+           "SET SESSION jit = off;"
+           "END IF;"
+           "END $$;"
+           (when expected-schema
+             (block-on-schema-mismatch expected-schema))]
+          (str/join " ")
+          (.setConnectionInitSql config))
+     (when rewrite-batched-inserts
+       (.setProperty (.getDataSourceProperties config) "reWriteBatchedInserts" rewrite-batched-inserts))
      (some->> pool-name (.setPoolName config))
      (some->> connection-timeout (.setConnectionTimeout config))
      (some->> maximum-pool-size (.setMaximumPoolSize config))
@@ -661,19 +709,20 @@
 (defn current-database []
   (-> "select current_database();" query-to-vec first :current_database))
 
-(defn-validated has-database-privilege?
-  [user db privilege] :- s/Bool
+(defn-validated has-database-privilege? :- s/Bool
+  [user db privilege]
   (-> ["select has_database_privilege(?, ?, ?)" user db privilege]
       query-to-vec first :has_database_privilege))
 
-(defn-validated has-role?
-  [user role privilege] :- s/Bool
+(defn-validated has-role? :- s/Bool
+  [user role privilege]
   (-> ["select pg_has_role(?, ?, ?)" user role privilege]
       query-to-vec first :pg_has_role))
 
-(defn disconnect-db [db]
+(defn disconnect-db
   "Forcibly disconnects all connections to the named db.  Requires
   that the current DB session has sufficient authorization."
+  [db]
   (query-to-vec
    (str "select pg_terminate_backend (pg_stat_activity.pid)"
         "  from pg_stat_activity"
@@ -681,10 +730,11 @@
         "    and pid <> pg_backend_pid()")
    db))
 
-(defn disconnect-db-role [db user]
+(defn disconnect-db-role
   "Forcibly disconnects all connections from the specified role to the
   named db.  Requires that the current DB session has sufficient
   authorization."
+  [db user]
   (query-to-vec
    (str "select pg_terminate_backend (pg_stat_activity.pid)"
         "  from pg_stat_activity"
