@@ -1,12 +1,10 @@
 (ns puppetlabs.trapperkeeper.internal
-  (:refer-clojure :exclude [boolean?])
   (:import (clojure.lang ExceptionInfo IFn IDeref)
            (java.lang ArithmeticException NumberFormatException))
   (:require [clojure.tools.logging :as log]
             [beckon]
             [plumbing.graph :as graph]
             [slingshot.slingshot :refer [throw+]]
-            [puppetlabs.kitchensink.core :refer [add-shutdown-hook! boolean? cli!]]
             [puppetlabs.trapperkeeper.config :refer [config-service get-in-config]]
             [puppetlabs.trapperkeeper.app :as a]
             [puppetlabs.trapperkeeper.common :as common]
@@ -176,7 +174,7 @@
                      ["-r" "--restart-file RESTART-FILE"
                       (i18n/trs "Path to a file whose contents are incremented each time all of the configured services have been started.")]]
         required    []]
-    (first (cli! cli-args specs required))))
+    (first (ks/cli! cli-args specs required))))
 
 (schema/defn ^:always-validate run-lifecycle-fn!
   "Run a lifecycle function for a service.  Required arguments:
@@ -302,7 +300,7 @@
 (defn restart-tk-apps
   "Call restart on all tk apps."
   [apps]
-  (log/debug (i18n/trs "SIGHUP handler restarting TK apps."))
+  (log/info (i18n/trs "SIGHUP handler restarting TK apps."))
   (doseq [app apps]
     (let [{:keys [lifecycle-channel]} @(a/app-context app)
           restart-fn #(a/restart app)]
@@ -347,19 +345,35 @@
 ;;;; regarding the cause of the shutdown, and is intended to be passed back
 ;;;; in to the top-level functions that perform various shutdown steps.
 
+(def exit-request-schema
+  "A process exit request like
+  {:status 7
+   :messages [[\"something for stderr\n\" *err*]]
+              [\"something for stdout\n\" *out*]]
+              [\"something else for stderr\n\" *err*]]"
+  {:status schema/Int
+   :messages [[(schema/one schema/Str "message")
+               (schema/one java.io.Writer "stream")]]})
+
 (def ^{:private true
        :doc "The possible causes for shutdown to be initiated."}
   shutdown-causes #{:requested :service-error :jvm-shutdown-hook})
 
-(defn request-shutdown*
+(schema/defn request-shutdown*
   "Initiate the normal shutdown of TrapperKeeper. This is asynchronous.
-  It is assumed that `wait-for-app-shutdown` has been called and is blocking.
-  Intended to be used by application services (likely their worker threads)
-  to programatically trigger application shutdown.
-  Note that this is exposed via the `shutdown-service` where it is a no-arg
-  function."
-  [shutdown-reason-promise]
-  (deliver shutdown-reason-promise {:cause :requested}))
+  It is assumed that `wait-for-app-shutdown` has been called and is
+  blocking.  Intended to be used by application services (likely their
+  worker threads) to programatically trigger application shutdown.
+  Note that this is exposed via the `shutdown-service`.  See the
+  protocol documentation for addiitonal information."
+  ([shutdown-reason-promise]
+   (request-shutdown* shutdown-reason-promise {}))
+  ([shutdown-reason-promise
+    opts :- {(schema/optional-key :puppetlabs.trapperkeeper.core/exit)
+             exit-request-schema}]
+   (deliver shutdown-reason-promise
+            (merge {:cause :requested}
+                   (select-keys opts [:puppetlabs.trapperkeeper.core/exit])))))
 
 (defn shutdown-on-error*
   "A higher-order function that is intended to be used as a wrapper around
@@ -397,7 +411,11 @@
 (defprotocol ShutdownService
   (get-shutdown-reason [this])
   (wait-for-shutdown [this])
-  (request-shutdown [this] "Asynchronously trigger normal shutdown")
+  (request-shutdown [this] [this opts]
+    "Asynchronously triggers normal shutdown.  A specific exit status
+    and final messages to display can be requested by associating an
+    exit-request-schema compatible value with
+    the :puppetlabs.trapperkeeper.core/exit key in the opts map.")
   (shutdown-on-error [this svc-id f] [this svc-id f on-error]
     "Higher-order function to execute application logic and trigger shutdown in
     the event of an exception"))
@@ -426,27 +444,40 @@
                                   @shutdown-reason-promise))
     (wait-for-shutdown [this] (deref shutdown-reason-promise))
     (request-shutdown [this]  (request-shutdown* shutdown-reason-promise))
+    (request-shutdown [this opts]  (request-shutdown* shutdown-reason-promise opts))
     (shutdown-on-error [this svc-id f] (shutdown-on-error* shutdown-reason-promise app-context svc-id f))
     (shutdown-on-error [this svc-id f on-error] (shutdown-on-error* shutdown-reason-promise app-context svc-id f on-error))))
 
-(schema/defn ^:always-validate shutdown!
+(schema/defn ^:always-validate shutdown! :- [Throwable]
   "Perform shutdown calling the `stop` lifecycle function on each service,
-   in reverse order (to account for dependency relationships)."
+   in reverse order (to account for dependency relationships).
+   Returns collection of exceptions thrown during shutdown sequence execution."
   [app-context :- (schema/atom a/TrapperkeeperAppContext)]
   (log/info (i18n/trs "Beginning shutdown sequence"))
   (let [{:keys [ordered-services shutdown-channel lifecycle-worker]} @app-context
-        shutdown-fn (fn [] (doseq [[service-id s] (reverse ordered-services)]
-                             (try
-                               (run-lifecycle-fn! app-context s/stop "stop" service-id s)
-                               (catch Exception e
-                                 (log/error e (i18n/trs "Encountered error during shutdown sequence"))))))]
+        errors-chan (async/promise-chan)
+        shutdown-fn (fn [] (let [results
+                                 (doall
+                                  (keep
+                                   (fn [[service-id s]]
+                                     (try
+                                       (run-lifecycle-fn! app-context s/stop "stop" service-id s)
+                                       nil
+                                       (catch Exception e
+                                         (log/error e (i18n/trs "Encountered error during shutdown sequence"))
+                                         e)))
+                                   (reverse ordered-services)))]
+                             (async/put! errors-chan results)))]
     (log/trace (i18n/trs "Putting shutdown message on shutdown channel."))
     (async/>!! shutdown-channel {:type :shutdown
                                  :task-function shutdown-fn})
     ;; wait for the channel to send us the return value so we know it's done
     (log/trace (i18n/trs "Waiting for response to shutdown message from lifecycle worker."))
     (if (not (nil? (async/<!! lifecycle-worker)))
-      (log/info (i18n/trs "Finished shutdown sequence"))
+      (do
+        (log/info (i18n/trs "Finished shutdown sequence"))
+        ;; deliver errors from stopped services if any
+        (async/<!! errors-chan))
       ;; else, the read from the channel returned a nil because it was closed,
       ;; indicating that there was already a shutdown in progress, and thus the
       ;; redundant shutdown request was ignored
@@ -458,7 +489,7 @@
    shutdown-reason-promise :- IDeref]
   (let [shutdown-service        (shutdown-service shutdown-reason-promise
                                                   app-context)]
-    (add-shutdown-hook! (fn []
+    (ks/add-shutdown-hook! (fn []
                           (when-not (realized? shutdown-reason-promise)
                             (log/info (i18n/trs "Shutting down due to JVM shutdown hook."))
                             (shutdown! app-context)
@@ -512,7 +543,7 @@
   to rely on control returning to the main thread where it can perform shutdown."
   [shutdown-reason]
   {:pre  [(map? shutdown-reason)]
-   :post [(boolean? %)]}
+   :post [(ks/boolean? %)]}
   (contains? (disj shutdown-causes :jvm-shutdown-hook) (:cause shutdown-reason)))
 
 (defn call-error-handler!
@@ -587,8 +618,14 @@
         (inc-restart-counter! this)
         this)
       (a/stop [this]
-        (shutdown! app-context)
-        this)
+        (a/stop this false))
+      (a/stop [this throw?]
+        (let [errors (shutdown! app-context)]
+          (if (and throw? (seq errors))
+            (let [msg (i18n/trs "Error during app shutdown!")]
+              (log/error msg)
+              (throw (ex-info msg {:errors errors})))
+            this)))
       (a/restart [this]
         (try
           (run-lifecycle-fns app-context s/stop "stop" (reverse ordered-services))
